/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.
*/
#ifndef PRCORE_ARRAY_HPP
#define PRCORE_ARRAY_HPP


#include <assert.h>


namespace prcore
{

	template <class T>
	class Array
	{
		public:

		Array();
		Array(int size);
		Array(const Array& array);
		~Array();

		void		operator = (const Array& array);
		T&			operator [] (int index);
		const T&	operator [] (int index) const;

		bool		IsEmpty() const;
		int			GetSize() const;
		const T*	GetArray() const;
		T*			GetArray();

		void		SetSize(int size, bool preserve=false);
		T&			PushBack(const T& item);
		T&			PushBack();
		T&			PopBack();
		void		Remove(const T& item);
		void		RemoveBlock(int index, int count);
		void		ResetIndex();
		void		TrimArray();
		void		Clear();

		private:

		T*			mData;
		int			mMaxSize;
		int			mOffset;
	};


	// inline methods

	template <class T>
	Array<T>::Array()
	{
		mData = NULL;
		mMaxSize = 0;
		mOffset = 0;
	}


	template <class T>
	Array<T>::Array(int size)
	{
		mData = new T[size];
		mMaxSize = size;
		mOffset = size;
	}


	template <class T>
	Array<T>::Array(const Array& array)
	{
		int size = array.GetSize();
		if ( size )
		{
			mData = new T[size];
			mMaxSize = size;
			mOffset = size;

			for ( int i=0; i<size; i++ )
				mData[i] = array.mData[i];
		}
		else
		{
			mData = NULL;
			mMaxSize = 0;
			mOffset = 0;
		}
	}


	template <class T>
	Array<T>::~Array()
	{
		delete[] mData;
	}


	template <class T>
	void Array<T>::operator = (const Array& array)
	{
		int size = array.GetSize();
		SetSize(size,false);

		for ( int i=0; i<size; i++ )
			mData[i] = array.mData[i];
	}


	template <class T>
	T& Array<T>::operator [] (int index)
	{
		assert( index>=0 && index < mMaxSize );
		return mData[index];
	}


	template <class T>
	const T& Array<T>::operator [] (int index) const
	{
		assert( index>=0 && index < mMaxSize );
		return mData[index];
	}


	template <class T>
	bool Array<T>::IsEmpty() const
	{
		return mOffset ? false : true;
	}


	template <class T>
	int Array<T>::GetSize() const
	{
		return mOffset;
	}


	template <class T>
	const T* Array<T>::GetArray() const
	{
		return mData;
	}


	template <class T>
	T* Array<T>::GetArray()
	{
		return mData;
	}


	template <class T>
	void Array<T>::SetSize(int size, bool preserve)
	{
		if ( size == mMaxSize )
			return;

		T* array = NULL;

		if ( size )
		{
			array = new T[size];

			if ( mData && preserve )
			{
				int size0 = GetSize();
				int count = (size0 > size) ? size : size0;
				for ( int i=0; i<count; i++ )
					array[i] = mData[i];
			}
		}

		delete[] mData;
		mData = array;
		mMaxSize = size;
		mOffset = size;
	}


	template <class T>
	T& Array<T>::PushBack(const T& item)
	{
		if ( mOffset >= mMaxSize )
		{
			if ( !mMaxSize )
			{
				mData = new T[4];
				mMaxSize = 4;
				mOffset = 0;
			}
			else
			{
				int offset = mOffset;
				int size = mMaxSize * 2;
				SetSize(size,true);
				mOffset = offset;
			}
		}
	
		T& ref = mData[mOffset++];
		ref = item;
		return ref;
	}


	template <class T>
	T& Array<T>::PushBack()
	{
		if ( mOffset >= mMaxSize )
		{
			if ( !mMaxSize )
			{
				mData = new T[4];
				mMaxSize = 4;
				mOffset = 0;
			}
			else
			{
				int offset = mOffset;
				int size = mMaxSize * 2;
				SetSize(size,true);
				mOffset = offset;
			}
		}

		return mData[mOffset++];
	}


	template <class T>
	T& Array<T>::PopBack()
	{
		assert( GetSize() > 0 );
		return mData[--mOffset];
	}


	template <class T>
	void Array<T>::Remove(const T& item)
	{
		int count = GetSize();
		T* src = mData;
		T* dest = mData;

		for ( int i=0; i<count; i++ )
		{
			if ( *src != item )
			{
				if ( src != dest )
					*dest = *src;

				dest++;
			}
			src++;
		}

		mOffset = (int)(dest - mData);
	}


	template <class T>
	void Array<T>::RemoveBlock(int index, int count)
	{
		T* dest = mData + index;
		T* src = mData + index + count;
		int left = (int)((mData+mOffset) - src);

		for ( int i=0; i<left; i++ )
			*dest++ = *src++;

		mOffset = (int)(dest - mData);
	}


	template <class T>
	void Array<T>::ResetIndex()
	{
		mOffset = 0;
	}


	template <class T>
	void Array<T>::TrimArray()
	{
		int size = GetSize();
		SetSize(size,true);
	}


	template <class T>
	void Array<T>::Clear()
	{
		SetSize(0,false);
	}

} // namespace prcore


#endif
