/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.
*/
#ifndef PRCORE_ENDIAN_HPP
#define PRCORE_ENDIAN_HPP


#include <prcore/configure.hpp>


namespace prcore
{

	// byte swaps
	
	inline void ByteSwap(int8&)
	{
		// satisfy macros defined later
	}

	inline void ByteSwap(uint8&)
	{
		// satisfy macros defined later
	}

	inline void ByteSwap(int16& i)
	{
		i = (i << 8) | (i >> 8);
	}

	inline void ByteSwap(uint16& i)
	{
		i = (i << 8) | (i >> 8);
	}

	inline void ByteSwap(int32& i)
	{
		i = (i >> 24) | (i >> 8) & 0x0000ff00 | (i << 8) & 0x00ff0000 | (i << 24);
	}

	inline void ByteSwap(uint32& i)
	{
		i = (i >> 24) | (i >> 8) & 0x0000ff00 | (i << 8) & 0x00ff0000 | (i << 24);
	}

	inline void ByteSwap(float& v)
	{
		uint32& i = (uint32&)v;
		i = (i >> 24) | (i >> 8) & 0x0000ff00 | (i << 8) & 0x00ff0000 | (i << 24);
	}

	inline void ByteSwap(int64& i)
	{
		uint32* p = (uint32*)&i;
		uint32 u = (p[0] >> 24) | (p[0] << 24) | ((p[0] & 0x00ff0000) >> 8) | ((p[0] & 0x0000ff00) << 8);
		p[0] = (p[1] >> 24) | (p[1] << 24) | ((p[1] & 0x00ff0000) >> 8) | ((p[1] & 0x0000ff00) << 8);
		p[1] = u;
	}

	inline void ByteSwap(uint64& i)
	{
		uint32* p = (uint32*)&i;
		uint32 u = (p[0] >> 24) | (p[0] << 24) | ((p[0] & 0x00ff0000) >> 8) | ((p[0] & 0x0000ff00) << 8);
		p[0] = (p[1] >> 24) | (p[1] << 24) | ((p[1] & 0x00ff0000) >> 8) | ((p[1] & 0x0000ff00) << 8);
		p[1] = u;
	}

	inline void ByteSwap(double& i)
	{
		uint32* p = (uint32*)&i;
		uint32 u = (p[0] >> 24) | (p[0] << 24) | ((p[0] & 0x00ff0000) >> 8) | ((p[0] & 0x0000ff00) << 8);
		p[0] = (p[1] >> 24) | (p[1] << 24) | ((p[1] & 0x00ff0000) >> 8) | ((p[1] & 0x0000ff00) << 8);
		p[1] = u;
	}

	// remap templates

	template <class T>
	inline T LittleEndianRemap(T v)
	{
		#ifdef PRCORE_BIG_ENDIAN
		ByteSwap( v );
		#endif
		return v;
	}

	template <class T>
	inline T BigEndianRemap(T v)
	{
		#ifdef PRCORE_LITTLE_ENDIAN
		ByteSwap( v );
		#endif
		return v;
	}
	
	// implementation note:
	//
	// we must do the following conversion char-by-char so that
	// machines with strict alignment rules will not generate
	// unaligned memory access exception.

	template <class T, class R>
	inline T ReadLittleEndianBuffer(R& buffer)
	{
		assert( sizeof(buffer[0]) == 1 );

		T v;
		char*& ps = reinterpret_cast<char*&>(buffer);

		#ifdef PRCORE_LITTLE_ENDIAN

		// little-to-little endian conversion:
		char* pv = reinterpret_cast<char*>(&v);
		int size = sizeof(T);
		while ( size-- )
			*pv++ = *ps++;

		#else

		// little-to-big endian conversion:
		char* pv = reinterpret_cast<char*>(&v) + sizeof(T);
		int size = sizeof(T);
		while ( size-- )
			*--pv = *ps++;

		#endif

		return v;
	}

	template <class T, class R>
	inline T ReadBigEndianBuffer(R& buffer)
	{
		assert( sizeof(buffer[0]) == 1 );

		T v;
		char*& ps = reinterpret_cast<char*&>(buffer);

		#ifdef PRCORE_LITTLE_ENDIAN

		// little-to-big endian conversion:
		char* pv = reinterpret_cast<char*>(&v) + sizeof(T);
		int size = sizeof(T);
		while ( size-- )
			*--pv = *ps++;

		#else

		// big-to-big endian conversion:
		char* pv = reinterpret_cast<char*>(&v);
		int size = sizeof(T);
		while ( size-- )
			*pv++ = *ps++;

		#endif

		return v;
	}

	// runtime endianess

	inline bool IsLittleEndian()
	{
		int v = 1;
		return *((char*)&v) != 0;
	}

	inline bool IsBigEndian()
	{
		int v = 1;
		return *((char*)&v) == 0;
	}

} // namespace prcore


#endif
