/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.
*/
#ifndef PRCORE_LINKEDLIST_HPP
#define PRCORE_LINKEDLIST_HPP


#include <assert.h>


namespace prcore
{

	template <class T>
	class LinkedList
	{
		public:

		struct Iterator
		{
			Iterator*	prev;
			Iterator*	next;
			T			data;

			operator const T& () const { return data; }
			operator T& () { return data; }
		};

		LinkedList();
		LinkedList(const LinkedList& otherlist);
		~LinkedList();

		void			operator = (const LinkedList& otherlist);
		const T&		operator [] (int index) const;
		T&				operator [] (int index);

		bool			IsEmpty() const;
		int				GetSize() const;
		Iterator*		Begin() const;
		Iterator*		End() const;

		T&				PushBack(const T& item);
		T&				PushFront(const T& item);
		T&				PushBack();
		T&				PushFront();
		void			InsertBefore(Iterator*, const T& item);
		void			InsertAfter(Iterator*, const T& item);
		void			PopBack();
		void			PopFront();
		void			Remove(Iterator* node);
		void			Remove(const T& item);
		void			Erase(Iterator* first, Iterator* last);
		void			Clear();

		private:

		int				mSize;
		Iterator*		mpHead;
		Iterator*		mpTail;
		
		// iterator block manager
		struct IteratorBlock
		{
			enum { blocksize = 16 };
			
			IteratorBlock*	next;
			Iterator		itb[blocksize];
		};

		IteratorBlock*	mpBlockCache;
		Iterator*		mpIteratorCache;
		
		Iterator*		GetIterator();
		void			ReleaseIterator(Iterator*);
		void			ClearIteratorCache();
		Iterator*		PushNodeBack();
		Iterator*		PushNodeFront();
	};


	// inline methods

	template <class T>
	typename LinkedList<T>::Iterator* LinkedList<T>::GetIterator()
	{
		if ( !mpIteratorCache )
		{
			// allocate new block of iterators
			IteratorBlock* block = new IteratorBlock();

			// initialize the iterators
			const int blocksize = IteratorBlock::blocksize - 1;
			for ( int i=0; i<blocksize; ++i )
				block->itb[i].next = block->itb + i + 1;
			block->itb[blocksize].next = NULL;
		
			// add the block to release chain
			block->next = mpBlockCache;
			mpBlockCache = block;
		
			// append the block iterators to the iterator cache
			mpIteratorCache = block->itb;
		}

		Iterator* node = mpIteratorCache;
		mpIteratorCache = mpIteratorCache->next;

		return node;	
	}


	template <class T> 
	void LinkedList<T>::ReleaseIterator(Iterator* node)
	{
		assert( node != NULL );

		node->next = mpIteratorCache;
		mpIteratorCache = node;
	}


	template <class T> 
	void LinkedList<T>::ClearIteratorCache()
	{
		IteratorBlock* block = mpBlockCache;
		while ( block )
		{
			IteratorBlock* next = block->next;
			delete block;
			block = next;
		}

		mSize = 0;
		mpHead = NULL;
		mpTail = NULL;
		mpBlockCache = NULL;
		mpIteratorCache = NULL;
	}


	template <class T>
	typename LinkedList<T>::Iterator* LinkedList<T>::PushNodeBack()
	{
		Iterator* node = GetIterator();
		++mSize;

		node->prev = mpTail;
		node->next = NULL;

		if ( mpTail )
			mpTail->next = node;
		else
			mpHead = node;

		mpTail = node;

		return node;
	}


	template <class T>
	typename LinkedList<T>::Iterator* LinkedList<T>::PushNodeFront()
	{
		Iterator* node = GetIterator();
		++mSize;

		node->prev = NULL;
		node->next = mpHead;

		if ( mpHead )
			mpHead->prev = node;
		else
			mpTail = node;

		mpHead = node;

		return node;
	}

	
	template <class T> 
	LinkedList<T>::LinkedList()
	{
		mSize = 0;
		mpHead = NULL;
		mpTail = NULL;
		mpBlockCache = NULL;
		mpIteratorCache = NULL;
	}


	template <class T>
	LinkedList<T>::LinkedList(const LinkedList& otherlist)
	{
		mSize = 0;
		mpHead = NULL;
		mpTail = NULL;
		mpBlockCache = NULL;
		mpIteratorCache = NULL;
		*this = otherlist;
	}


	template <class T>
	LinkedList<T>::~LinkedList()
	{
		ClearIteratorCache();
	}


	template <class T>
	void LinkedList<T>::operator = (const LinkedList& otherlist)
	{
		// clear
		ClearIteratorCache();

		// copy data
		Iterator* node = otherlist.Begin();
		while ( node )
		{
			PushBack(*node);
			node = node->next;
		}
	}


	template <class T>
	const T& LinkedList<T>::operator [] (int index) const
	{
		int size = GetSize();
		assert ( index>=0 && index < size );

		if ( index > size/2 )
		{
			Iterator* node = End();
			int count = size - index;
			while ( --count )
				node = node->prev;
			
			return *node;
		}

		Iterator* node = Begin();
		while ( index-- )
			node = node->next;

		return *node;
	}


	template <class T>
	T& LinkedList<T>::operator [] (int index)
	{
		int size = GetSize();
		assert ( index>=0 && index < size );

		if ( index > size/2 )
		{
			Iterator* node = End();
			int count = size - index;
			while ( --count )
				node = node->prev;
			
			return *node;
		}

		Iterator* node = Begin();
		while ( index-- )
			node = node->next;

		return *node;
	}


	template <class T>
	bool LinkedList<T>::IsEmpty() const
	{
		return GetSize() ? false : true;
	}


	template <class T>
	int LinkedList<T>::GetSize() const
	{
		return mSize;
	}


	template <class T>
	typename LinkedList<T>::Iterator* LinkedList<T>::Begin() const
	{
		return mpHead;
	}


	template <class T>
	typename LinkedList<T>::Iterator* LinkedList<T>::End() const
	{
		return mpTail;
	}


	template <class T>
	T& LinkedList<T>::PushBack(const T& item)
	{
		Iterator* node = PushNodeBack();
		node->data = item;
		return node->data;
	}


	template <class T>
	T& LinkedList<T>::PushFront(const T& item)
	{
		Iterator* node = PushNodeFront();
		node->data = item;
		return node->data;
	}


	template <class T>
	T& LinkedList<T>::PushBack()
	{
		Iterator* node = PushNodeBack();
		return node->data;
	}


	template <class T>
	T& LinkedList<T>::PushFront()
	{
		Iterator* node = PushNodeFront();
		return node->data;
	}


	template <class T>
	void LinkedList<T>::InsertBefore(Iterator* it, const T& item)
	{
		assert( it );
		Iterator* node = GetIterator();
		++mSize;

		Iterator* prev = it->prev;
		it->prev = node;

		if ( prev )
			prev->next = node;
		else
			mpHead = node;

		node->next = it;
		node->prev = prev;
		node->data = item;
	}


	template <class T>
	void LinkedList<T>::InsertAfter(Iterator* it, const T& item)
	{
		assert( it );
		Iterator* node = GetIterator();
		++mSize;

		Iterator* next = it->next;
		it->next = node;

		if ( next )
			next->prev = node;
		else
			mpTail = node;

		node->prev = it;
		node->next = next;
		node->data = item;
	}


	template <class T>
	void LinkedList<T>::PopBack()
	{
		Iterator* node = mpTail;

		if ( node )
		{
			Iterator* prev = node->prev;

			if ( prev )
				prev->next = NULL;
			else
				mpHead = NULL;

			mpTail = prev;
			ReleaseIterator( node );
			--mSize;
		}
	}


	template <class T>
	void LinkedList<T>::PopFront()
	{
		Iterator* node = mpHead;
		if ( node )
		{
			Iterator* next = node->next;

			if ( next )
				next->prev = NULL;
			else
				mpTail = NULL;

			mpHead = next;
			ReleaseIterator( node );
			--mSize;
		}
	}


	template <class T>
	void LinkedList<T>::Remove(Iterator* node)
	{
		assert(node != NULL);
	
		Iterator* next = node->next;
		Iterator* prev = node->prev;

		if ( node == mpHead )
			mpHead = next;

		if ( node == mpTail )
			mpTail = node->prev;

		if ( prev )
			prev->next = next;

		if ( next )
			next->prev = prev;

		ReleaseIterator( node );
		--mSize;
	}


	template <class T>
	void LinkedList<T>::Remove(const T& item)
	{
		Iterator* node = mpHead;
		while ( node )
		{
			Iterator* next = node->next;

			if ( node->data == item )
				Remove(node);

			node = next;
		}
	}


	template <class T>
	void LinkedList<T>::Erase(Iterator* first, Iterator* last)
	{
		if ( IsEmpty() || !first || !last )
			return;

		if ( first == mpHead )
			mpHead = last->next;

		if ( last == mpTail )
			mpTail = first->prev;

		Iterator* node = first;
		while ( node )
		{
			Iterator* next = node->next;
			Iterator* prev = node->prev;

			if ( prev )
				prev->next = next;

			if ( next )
				next->prev = prev;

			ReleaseIterator( node );
			--mSize;
		
			if ( node==last )
				break;
			
			node = next;
		}
	}


	template <class T>
	void LinkedList<T>::Clear()
	{
		ClearIteratorCache();
	}

} // namespace prcore


#endif
