/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source:
		bitmap implementation

	revision history:
		Dec/18/1998 - Jukka Liimatta - initial revision
		Jan/24/2001 - Jukka Liimatta - renaissance build
		Feb/21/2002 - Jukka Liimatta - created BitmapCodecManager
*/
#include <prcore/filename.hpp>
#include <prcore/filestream.hpp>
#include <prcore/bitmap.hpp>

using namespace prcore;


// ========================================
// Bitmap
// ========================================

Bitmap::Bitmap()
{
}


Bitmap::Bitmap(const Bitmap& src)
{
	*this = src;
}


Bitmap::Bitmap(const Surface& surface)
{
	width = surface.GetWidth();
	height = surface.GetHeight();
	format = surface.GetPixelFormat();
	pitch = width * format.GetBytes();
	image = new uint8[ height*pitch ];

	// copy image
	uint8* s = surface.GetImage();
	uint8* d = image;
	int y = height;
	do
	{
		memcpy(d,s,pitch);
		s += surface.GetPitch();
		d += pitch;
	} while (--y);
}


Bitmap::Bitmap(int width_, int height_, const PixelFormat& pxf)
{
	width = width_;
	height = height_;
	pitch = width * pxf.GetBytes();
	format = pxf;
	image = new uint8[ width*height*pxf.GetBytes() ];
}


Bitmap::Bitmap(const char* filename)
{
	LoadImage( filename );
}


Bitmap::~Bitmap()
{
	if ( image )
		delete[] image;
}


void Bitmap::operator = (const Bitmap& src)
{
	if ( src.image )
	{
		// reallocate image
		if ( pitch!=src.pitch || height!=src.height )
		{
			if ( image )
				delete[] image;
			image = new uint8[ src.pitch*src.height ];
		}

		// copy image
		uint8* s = src.image;
		uint8* d = image;
		int y = src.height;
		do
		{
			memcpy(d,s,src.pitch);
			s += src.pitch;
			d += src.pitch;
		} while (--y);
	}
	else
	{
		if ( image )
			delete[] image;
		image = NULL;
	}
	
	width = src.width;
	height = src.height;
	pitch = src.pitch;
	format = src.format;
}


Bitmap Bitmap::GetBitmap(const Rect& area) const
{
	// assert
	assert( image );

	// test area
	if ( area.width < 1 || 
		area.height < 1 || 
		area.x < 0 || 
		area.y < 0 || 
		(area.x+area.width) > width || 
		(area.y+area.height) > height ) 
		return *this;

	// clip
	Rect q;
	if ( !q.Intersect(GetRect(),area) )
		return *this;

	// new bitmap
	Bitmap p(q.width,q.height,format);

	// copy image
	int bpp = format.GetBytes();
	uint8* s = image + q.y * pitch + q.x * bpp;
	uint8* d = p.image;

	for ( int y=0; y<q.height; y++ )
	{
		memcpy(d,s,p.pitch);
		s += pitch;
		d += p.pitch;
	}

	return p;
}


void Bitmap::LoadImage(const char* filename)
{
	// assert
	assert(filename);
	
	// find codec
	const BitmapCodec* codec = BitmapCodecManager::manager.FindInterface(GetFilenameEXT(filename));
	if ( codec )
	{
		if ( !codec->decode )
			return;

		// create stream
		FileStream stream(filename,Stream::READ);
		
		// decode
		codec->decode(*this,stream);
		
		// set interface name
		SetName(filename);
	}
}


void Bitmap::SetImage(int width_, int height_, const PixelFormat& pxf, void* image_)
{
	if ( image )
	delete[] image;
	
	width = width_;
	height = height_;
	pitch = width * pxf.GetBytes();
	format = pxf;
	image = (uint8*)image_;
}


void Bitmap::ResizeImage(int width_, int height_, bool filter)
{
	// assert
	assert( image );
	if ( width == width_ && height == height_ )	
		return;

	// choose blitmode
	BlitMode mode = filter ? BLIT_BILINEAR_SCALE : BLIT_SCALE;

	// conversion
	Bitmap temp(width_,height_,format);
	temp.BlitImage(*this,mode);
	
	// swap pointers
	if ( image )
		delete[] image;
	image = temp.image;
	temp.image = NULL;
	
	format = temp.format;
	width = temp.width;
	height = temp.height;
	pitch = temp.pitch;
}


void Bitmap::ReformatImage(const PixelFormat& pxf)
{
	assert( image );
	if ( format==pxf )
		return;

	// indexed reformat means color quantization
	if ( pxf.IsIndexed() )
	{
		QuantizeImage();
		return;
	}

	// conversion
	Bitmap temp(width,height,pxf);
	temp.BlitImage(*this,BLIT_COPY);
	
	// swap pointers
	delete[] image;
	image = temp.image;
	temp.image = NULL;
	
	pitch = temp.pitch;
	format = pxf;
}


// ========================================
// BitmapCodecManager
// ========================================

	// the manager instance
	BitmapCodecManager BitmapCodecManager::manager;

	// forward declarations
	BitmapCodec CreateCodecTGA();
	BitmapCodec CreateCodecPCX();
	BitmapCodec CreateCodecRGB();
	BitmapCodec CreateCodecBMP();
	BitmapCodec CreateCodecJPG();
	BitmapCodec CreateCodecPNG();
	BitmapCodec CreateCodecDDS();


BitmapCodecManager::BitmapCodecManager()
{
	manager.RegisterInterface(CreateCodecTGA());
	manager.RegisterInterface(CreateCodecPCX());
	manager.RegisterInterface(CreateCodecRGB());
	manager.RegisterInterface(CreateCodecBMP());
	manager.RegisterInterface(CreateCodecJPG());
	manager.RegisterInterface(CreateCodecPNG());
	manager.RegisterInterface(CreateCodecDDS());
}



