/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source: 
		BitmapCodec "bmp"

	revision history:
		Dec/02/2000 - Jukka Liimatta - initial revision
		Jan/24/2001 - Jukka Liimatta - renaissance build
		Feb/13/2002 - Jukka Liimatta - added encoding support
		Feb/21/2002 - Jukka Liimatta - rewrote codec interface

	todo:
	- 4bit win32 rle decoder
	- topdown images (negative height)
*/
#include <prcore/endian.hpp>
#include <prcore/bitmap.hpp>

using namespace prcore;


// =================================================
// codec
// =================================================

static bool decode(Bitmap& target, Stream& stream)
{
	// reset stream
	stream.Seek(0,Stream::START);

	// read header
	uint16 signature = ReadLittleEndian<uint16>(stream);
	if ( signature != PRCORE_CODE16('B','M') )
		return false;

	ReadLittleEndian<uint32>(stream); // filesize
	ReadLittleEndian<uint32>(stream); // reserved
	uint32 dataoffset = ReadLittleEndian<uint32>(stream);

	uint32 size = ReadLittleEndian<uint32>(stream);
	uint32 width = 0;
	uint32 height = 0;
	//uint16 planes = 0;
	uint16 bitcount = 0;

	// compression
	enum bi_type
	{
		bi_rgb = 0,
		bi_rle4 = 1,
		bi_rle8 = 2,
		bi_bitfields = 3
	};
	uint32 compression = bi_rgb;

	bool OS2 = false;
	if ( size == 40 )
	{
		// win32 bmp
		width = ReadLittleEndian<uint32>(stream);
		height = ReadLittleEndian<uint32>(stream);
		/*planes = */ReadLittleEndian<uint16>(stream);
		bitcount = ReadLittleEndian<uint16>(stream);
		compression = ReadLittleEndian<uint32>(stream);

		ReadLittleEndian<uint32>(stream); // imagesize
		ReadLittleEndian<uint32>(stream); // xres
		ReadLittleEndian<uint32>(stream); // yres
		ReadLittleEndian<uint32>(stream); // colorused
		ReadLittleEndian<uint32>(stream); // colorimportant
	}
	else if ( (size == 12) || (size == 240) )
	{
		// OS/2 bmp
		OS2 = true;
		width = ReadLittleEndian<uint16>(stream);
		height = ReadLittleEndian<uint16>(stream);
		/*planes = */ReadLittleEndian<uint16>(stream);
		bitcount = ReadLittleEndian<uint16>(stream);
	}
	else
	{
		return false;
	}

	int palsize = 0;
	int pitch = 0;

	// pixelformat
	PixelFormat pxf;
	switch ( bitcount )
	{
		case 1:
			pxf = PixelFormat(PALETTE8(NULL));
			palsize = 2;
			pitch = (width+7) >> 3;
			break;

		case 4:
			pxf = PixelFormat(PALETTE8(NULL));
			palsize = 16;
			pitch = (width+3) >> 2;
			break;

		case 8:
			pxf = PixelFormat(PALETTE8(NULL));
			palsize = 256;
			pitch = width;
			break;

		case 16:
			pxf = PixelFormat(RGB565);
			palsize = 0;
			pitch = width * 2;
			break;

		case 24:
			pxf = PixelFormat(RGB888);
			palsize = 0;
			pitch = width * 3;
			break;

		case 32:
			pxf = PixelFormat(ARGB8888);
			palsize = 0;
			pitch = width * 4;
			break;

		default:
			return false;
	}

	// read colormap
	Color32* palette = pxf.GetPalette();
	for ( int i=0; i<palsize; i++ )
	{
		Color32 color;
		color.b = ReadLittleEndian<uint8>(stream);
		color.g = ReadLittleEndian<uint8>(stream);
		color.r = ReadLittleEndian<uint8>(stream);
		color.a = OS2 ? 0xff : ReadLittleEndian<uint8>(stream);

		if ( palette )
			*palette++ = color;
	}

	// 16bit and 32bit color masks ( optional field in the stream )
	if ( (compression!=bi_rgb) && (bitcount==16 || bitcount==32) )
	{
		uint32 rmask = ReadLittleEndian<uint32>(stream);
		uint32 gmask = ReadLittleEndian<uint32>(stream);
		uint32 bmask = ReadLittleEndian<uint32>(stream);
		uint32 amask = ~(rmask|gmask|bmask);

		pxf = PixelFormat(bitcount,rmask,gmask,bmask,amask);
	}

	// initialize decode target
	target = Bitmap(width,height,pxf);

	// 32bit alignment adjust
	int align = ((pitch+3) & 0xfffffffc) - pitch;

	// seek to image
	stream.Seek(dataoffset,Stream::START);

	// read image
	if ( compression != bi_rgb )
	{
		// ================================
		// compression: RLE
		// ================================
		switch ( bitcount )
		{
			case 4:
			{
				//TODO: write support for 4bit RLE compressed bmp's
				break;
			}

			case 8:
			{
				memset( target.GetImage(), 0, width*height );
				int y = (height-1) * 2 - 1;
				while ( y > 0 )
				{
					int x = 0;
					while ( x < (int)width )
					{
						uint16 data;
						stream.Read(&data,sizeof(data));
						uint8 n = data & 0xff;
						uint8 c = data >> 8;

						if ( n )
						{
							char* buffer = (char*)target.GetImage() + pitch*(y/2) + x;
							for ( int i=0; i<n; i++ )
								*buffer++ = c;
							x += n;
						}
						else
						{
							switch ( c )
							{
								case 0:
									x = width;
									break;
		
								case 1:
									y = 0;
									x = width;
									break;

								case 2:
								{
									uint16 delta;
									stream.Read(&delta,sizeof(delta));
									x += (delta & 0xff);
									y -= (delta >> 8);
									break;
								}

								default:
								{
									char* buffer = (char*)target.GetImage() + pitch*(y/2) + x;
									stream.Read(buffer,c);
									x += c;

									if ( c & 1 )
										stream.Seek(1,Stream::CURRENT);
									break;
								}
							}
						}
					}
					--y;
				}
				break;
			}

			default:
				return false;
		}
	}
	else
	{
		// ================================
		// compression: LINEAR
		// ================================
		switch ( bitcount )
		{
			case 1:
			case 4:
			{
				for ( int y=0; y<(int)height; y++ )
				{
					char* dest = (char*)target.GetImage() + (height-y-1)*target.GetWidth();

					int bits = bitcount;
					uint32 data = 0;
					uint32 mask = (1 << bitcount) - 1;
					for ( int x=0; x<(int)width; x++ )
					{
						bits -= bitcount;
						if ( !bits )
						{
							stream.Read(&data,sizeof(data));
							ByteSwap(data);
							bits = 32;
						}
						*dest++ = (data>>(bits-bitcount)) & mask;
					}
				}
				break;
			}

			case 8:
			case 16:
			case 24:
			case 32:
			{
				for ( int y=0; y<(int)height; y++ )
				{
					char* dest = (char*)target.GetImage() + (height-y-1)*pitch;
					stream.Read(dest,pitch);

					if ( align )
						stream.Seek(align,Stream::CURRENT);
				}
				break;
			}

			default:
				return false;
		}
	}

	return true;
}


static bool encode(Stream& stream, Surface& surface)
{

	// info
	uint32 width = surface.GetWidth();
	uint32 height = surface.GetHeight();
	uint16 planes = 1;
	uint16 bitcount = 8;
	uint32 compression = 0;
	uint32 dataoffset = 18 + 16 + 20;

	// choose pixelformat
	PixelFormat pxf = surface.GetPixelFormat();
	if ( pxf.IsIndexed() )
	{
		dataoffset += 1024;
	}
	else
	{
		switch ( pxf.GetBits() )
		{
			case 8: // flow to 16bit format
			case 16: pxf = PixelFormat(RGB565); planes = 3; bitcount = 16; break;
			case 24: pxf = PixelFormat(RGB888); planes = 3; bitcount = 24; break;
			case 32: pxf = PixelFormat(ARGB8888); planes = 4; bitcount = 32; break;
		}
	}

	int pitch = width * pxf.GetBytes();
	int align = ((pitch + 3) & 0xfffffffc) - pitch;
	uint32 filesize = dataoffset + (height * (pitch+align));

	// write header
	// 18 bytes
	WriteLittleEndian<uint16>(stream,PRCORE_CODE16('B','M')); // id
	WriteLittleEndian<uint32>(stream,filesize); // filesize
	WriteLittleEndian<uint32>(stream,0); // reserved
	WriteLittleEndian<uint32>(stream,dataoffset); // data offset
	WriteLittleEndian<uint32>(stream,40); // header size

	// 16 bytes
	WriteLittleEndian<uint32>(stream,width);
	WriteLittleEndian<uint32>(stream,height);
	WriteLittleEndian<uint16>(stream,planes);
	WriteLittleEndian<uint16>(stream,bitcount);
	WriteLittleEndian<uint32>(stream,compression);

	// 20 bytes
	WriteLittleEndian<uint32>(stream,0); // imagesize
	WriteLittleEndian<uint32>(stream,0); // xres
	WriteLittleEndian<uint32>(stream,0); // yres
	WriteLittleEndian<uint32>(stream,0); // colorused
	WriteLittleEndian<uint32>(stream,0); // colorimportant
	
	// write colormap
	if ( pxf.IsIndexed() )
	{
		Color32* palette = pxf.GetPalette();
		for ( int i=0; i<256; ++i )
		{
			WriteLittleEndian<uint8>(stream,palette[i].b);
			WriteLittleEndian<uint8>(stream,palette[i].g);
			WriteLittleEndian<uint8>(stream,palette[i].r);
			WriteLittleEndian<uint8>(stream,palette[i].a);
		}
	}
	
	// write image
	Bitmap encobj(width,height,pxf);
	encobj.BlitImage(surface);
	
	uint8* image = (uint8*)encobj.GetImage();
	
	for ( int y=0; y<(int)height; y++ )
	{
		uint8* ptr = image + (height-y-1) * pitch;
		stream.Write(ptr,pitch);
		
		for ( int j=0; j<align; ++j )
			WriteLittleEndian<uint8>(stream,0);
	}

	return true;
}


// =================================================
// factory
// =================================================

BitmapCodec CreateCodecBMP()
{
	static const char* ext[] = {"bmp"};
	BitmapCodec codec;

	codec.count = sizeof(ext)/sizeof(ext[0]);
	codec.extension = ext;
	codec.decode = decode;
	codec.encode = encode;

	return codec;
}



