/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source:
		filestream implementation

	revision history:
		Jan/24/2001 - Jukka Liimatta - initial revision/renaissance build
		Jan/30/2001 - Timo Saarinen - read cache
		Feb/01/2001 - Jukka Liimatta - filefactory
		Oct/22/2001 - Jukka Liimatta - added stristr() to make detecting filesystem non-casesensitive
		Feb/08/2002 - Francesco Montecuccoli - fixed error in file(-system) extension detection
		Feb/15/2002 - Jukka Liimatta - fixed write seek bug (writes don't go through cache, so seek must be forwarded to the actual filesystem plugin)
		Feb/23/2002 - Jukka Liimatta - created FileCodecManager
		Feb/23/2002 - Francesco Montecuccoli - created FileIndex
		Mar/03/2002 - Francesco Montecuccoli - added support for nested file codecs
*/
#include <prcore/exception.hpp>
#include <prcore/linkedlist.hpp>
#include <prcore/filestream.hpp>

using namespace prcore;


// ========================================
// stristr()
// ========================================

static const char* stristr(const char* string, const char* ext)
{
	for ( ;; )
	{
		const char* s = string;

		// substring search loop
		for ( const char* e=ext; *e; ++e,++s )
		{
			if ( !*s )
				return NULL;

			if ( tolower(*s) != tolower(*e) )
			{
				// the substring is not same, abort search
				s = NULL;
				break;
			}
		}

		// the substring search loop succeeded
		if ( s )
			return string;

		++string;
	}
}


// ========================================
// global strings
// ========================================

static String password;
static String basepath;


// ========================================
// FileCodecManager
// ========================================

	// the manager instance
	FileCodecManager FileCodecManager::manager;

	// forward declarations
	FileCodec CreateCodecFILE();
	FileCodec CreateCodecZIP();


FileCodecManager::FileCodecManager()
{
	manager.RegisterInterface(CreateCodecZIP());
}



// ========================================
// FindFileCodec
// ========================================

namespace prcore
{
	// forward declaration to OS dependant function
	bool IsFolder(const char* path);
}


static const FileCodec* FindFileCodec(Stream* archive, const FileCodec* current, const String& fullpath, int& pos)
{
	// iterate looking for a filefactory, last registered are checked first
	for ( int i=FileCodecManager::manager.GetInterfaceCount()-1; i>=0; --i )
	{
		const FileCodec* codec = FileCodecManager::manager.GetInterface(i);

		for ( int j=0; j<codec->count; ++j )
		{
			// next supported extension
			String ext;
			ext << "." << codec->extension[j];

			// find extension as substring of filename
			const char* p = stristr(fullpath,ext);
			
			// be sure you actually got the extension
			while ( p )
			{
				// found extension, skip it to get filename
				p += ext.GetLength();

				// filename must be prefixed with slash
				if ( *p=='/' || *p=='\\' )
				{
					// pos is where the archive name ends
					// and filename starts...
					pos = (int)(p-(const char*)fullpath);

					// be sure we have not found a real folder
					if ( archive )
						archive->AddRef(); // don't want to release this stream
					if ( !current->isfolder(archive,fullpath.SubString(0,pos)) )
					{
						// mission impossible complete
						return codec;
					}
				}

				// try again...
				p = stristr(p,ext);
			}
		}
	}

	// not found any other codecs, keep current
	pos = -1;
	return current;
}


// WARNING: parameter filename can be modified to represent the path and file name
// relative to the archive (that's why it is not const)
static const FileCodec RecursiveFindFileCodec(Stream*& archive, String& filename)
{
	// find appropriate codec for the filename
	int pos = 0;
	FileCodec defcodec = CreateCodecFILE();
	FileCodec codec = *FindFileCodec(archive,&defcodec,filename,pos);

	// cycle through inner codecs
	// pos == -1 means we have not found any other codecs
	while ( pos != -1 )
	{
		archive = defcodec.stream(archive,filename.SubString(0,pos),Stream::READ,password);
		if ( !archive )
		{
			String error("cannot access file: ");
			error += filename;
			PRCORE_EXCEPTION("FileStream::FileStream()",error);
		}
		defcodec = codec;
		filename = String((char*)filename+(pos+1));
		codec = *FindFileCodec(archive,&defcodec,filename,pos);
	}
	return codec;
}


// ========================================
// FileStream
// ========================================

FileStream::FileStream(const char* filename_user, AccessMode mode)
: stream(NULL),offset(0),page(-1),iscache(false)
{
	// read cache?
	if ( mode == Stream::READ )
		iscache = true;

	// create filename at basepath
	String filename(basepath);
	filename += filename_user;

	// find appropriate codec for the filename
	Stream* archive = NULL;
	const FileCodec codec = RecursiveFindFileCodec(archive,filename);

	// finally create stream
	stream = codec.stream(archive,filename,mode,password);
	if ( !stream )
	{
		String error("cannot access file: ");
		error += filename;
		PRCORE_EXCEPTION("FileStream::FileStream()",error);
	}
}


FileStream::~FileStream()
{
	if ( stream )
		delete stream;
}


int FileStream::Seek(int delta, SeekMode mode)
{
	if ( stream )
	{
		switch ( mode )
		{
			case START:		offset = delta; break;
			case CURRENT:	offset += delta; break;
			case END:		offset = stream->GetSize() - delta; break;
		}
		
		if ( !iscache )
			stream->Seek(offset,Stream::START);
	}
	return offset;
}


int FileStream::Read(void* target, int bytes)
{
	int offset0 = offset;
	if ( stream )
	{
		if ( bytes > CACHE_SIZE )
		{
			stream->Seek(offset,START);
			stream->Read(target,bytes);
			offset += bytes;
			page = -1; // invalidate cache after blockread
		}
		else
		{
			char* cbuffer = (char*)target;
			int cpos = offset;
			int cpage = cpos >> CACHE_BITS;

			// read page, if not cached
			if ( cpage != page )
			{
				stream->Seek(cpage<<CACHE_BITS,START);
				stream->Read(cache,CACHE_SIZE);
				page = cpage;
			}

			// cache boundry cross
			int bytes_left = CACHE_SIZE - (cpos & (CACHE_SIZE-1));
			while ( bytes > bytes_left )
			{
				memcpy(cbuffer,cache+(cpos & (CACHE_SIZE-1)),bytes_left);
				cpos += bytes_left;
				cbuffer += bytes_left;
				bytes -= bytes_left;

				stream->Seek(cpos,START);
				stream->Read(cache,CACHE_SIZE);
				page = cpos >> CACHE_BITS;
				bytes_left = CACHE_SIZE;
			}

			// leftovers
			memcpy(cbuffer,cache+(cpos & (CACHE_SIZE-1)),bytes);
			offset = cpos + bytes;
		}
	}
	return offset - offset0;
}


int FileStream::Write(const void* source, int bytes)
{
	int delta = stream ? stream->Write(source,bytes) : 0;
	offset += delta;
	return delta;
}


int FileStream::GetOffset() const
{
	return offset;
}


int FileStream::GetSize() const
{
	return stream ? 
		stream->GetSize() : 0;
}


bool FileStream::IsOpen() const
{
	return stream ?
		stream->IsOpen() : false;
}


bool FileStream::IsEOS() const
{
	int size = stream ? stream->GetSize() : 0;
	return offset >= size;
}


void FileStream::SetBasePath(const char* path)
{
	basepath = path;
}


void FileStream::SetPassword(const char* pass)
{
	password = pass;
}


String FileStream::GetBasePath()
{
	return basepath;
}


// =================================================
// FileIndex
// =================================================

FileIndex::FileIndex()
{
	CreateIndex("");
}


FileIndex::FileIndex(const char* userpath)
{
	CreateIndex(userpath);
}


FileIndex::~FileIndex()
{
}


void FileIndex::CreateIndex(const char* userpath)
{
	// apply basepath to userpath
	String path(basepath);
	path += userpath;

	// reset index
	mFileArray.ResetIndex();
	mFolderArray.ResetIndex();

	// find appropriate codec for the filename
	Stream* archive = NULL;
	const FileCodec codec = RecursiveFindFileCodec(archive,path);
	
	// finally access index
	codec.index(archive,path,mFileArray,mFolderArray);
}


int FileIndex::GetFileCount() const
{
	return mFileArray.GetSize();
}


int FileIndex::GetFolderCount() const
{
	return mFolderArray.GetSize();
}


const char* FileIndex::GetFileName(int index) const
{
	return mFileArray[index];
}


const char* FileIndex::GetFolderName(int index) const
{
	return mFolderArray[index];
}

