/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source:
		WindowGL / OpenGL Linux implementation
		
	revision history:
		???/??/2001 - Edmond Cote - initial revision
*/

#include <prcore/prcore.hpp>

#include <GL/gl.h>
#include <GL/glx.h>
#include <GL/glext.h> 
#include <X11/extensions/xf86vmode.h>

using namespace prcore;

struct ContextGLX
{
	GLXContext glXContext;
};

static Array<ModeGL> ModeList;
static XF86VidModeModeInfo **modes; 
static bool isFullscreen;

static bool EnumModeList(struct WinHandle mWinHandle)
{
  int major, minor; // XF86 Version Numbers
  int event_base, error_base;
  int i,n;
  
  // check if extension is availible!
  
  if(XF86VidModeQueryVersion(mWinHandle.dpy,&minor, &major) && \
     XF86VidModeQueryExtension(mWinHandle.dpy,&event_base,&error_base))
    {
      
      // query all modes

      XF86VidModeGetAllModeLines(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy),
				 &n, &modes);
      
      for(i=0;i<n;i++)
	{
	  ModeGL mode;
	  mode.width=modes[i]->hdisplay;
	  mode.height=modes[i]->vdisplay;
	  mode.bits=0;
	  mode.frequency=0;
	  
	  ModeList.PushBack(mode);
	}
      return true;
    }	
  return false;

}


WindowGL::WindowGL()
{
  isFullscreen=false;
}


WindowGL::~WindowGL()
{
  RestoreMode();
}


int WindowGL::GetModeCount() const
{
  return ModeList.GetSize();
}


ModeGL WindowGL::GetMode(int index) const
{
  assert( index>=0 && index<= ModeList.GetSize() );
  return ModeList[index];
}


bool WindowGL::SetMode(const ModeGL& mode)
{
  int i=0;
  
  // select availible video mode
  while(mode.width!=ModeList[i].width && mode.height!=ModeList[i].height)
    i++;
  
  // set video mode
  if(i >= 0 && i <= ModeList.GetSize() )
    {
      
      XMoveWindow(mWinHandle.dpy, mWinHandle.win, 0,0);
      XResizeWindow(mWinHandle.dpy, mWinHandle.win,mode.width,mode.height);
     
      XSetWindowBorderWidth(mWinHandle.dpy, mWinHandle.win,0);
      
      // Switch to Mode
      XF86VidModeSwitchToMode(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy),
			      modes[i]);
      // Set Viewport
      XF86VidModeSetViewPort(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy), 
			     0, 0);
      // Lock Video Mode
      XF86VidModeLockModeSwitch(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy),
				1); 
      

      // Grab Exclusive Keyboard / Mouse Access
      XGrabKeyboard (mWinHandle.dpy,mWinHandle.win, False, GrabModeAsync, 
		     GrabModeAsync, CurrentTime);
           
      XGrabPointer (mWinHandle.dpy, mWinHandle.win, True,
		    PointerMotionMask | ButtonPressMask | ButtonReleaseMask,
		    GrabModeAsync, GrabModeAsync, None, None, CurrentTime);
     
      // Hide Cursor
      HideCursor();
      
      isFullscreen = true;

      return true;
      
     
    }
  return false;
  
}

bool WindowGL::RestoreMode()
{
  int n;

  if(isFullscreen)
    {
      
      XF86VidModeGetAllModeLines(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy),
				 &n, &modes);
      
  
      // Unlock Video Mode
      XF86VidModeLockModeSwitch(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy),1);
      
      // Not sure about this.. but modes[0] should be the original mode
      XF86VidModeSwitchToMode(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy),
			      modes[0]);  
      
      // Reset Viewport
      XF86VidModeSetViewPort(mWinHandle.dpy, DefaultScreen(mWinHandle.dpy), 
			     0, 0);
      

  // Show Cursor
  ShowCursor();
    }
  
  return true;
}


ContextGL* WindowGL::CreateContext(InitGL* init)
{
  if(!mWinHandle.win)
    return NULL;
  
  if ( !ModeList.GetSize() )
    EnumModeList(mWinHandle);
  
  ContextGLX *context = new ContextGLX();
    
  // Create Rendering Context
  context->glXContext = glXCreateContext(mWinHandle.dpy, mWinHandle.VisualInfo,
					 NULL, True);

  if(context->glXContext == NULL)
    PRCORE_EXCEPTION("WindowGL::CreateContext", "Could not create rendering context!" );
  
  // set context
  //etContext( (ContextGL*)context );
  
  return (ContextGL*)context;

}


void WindowGL::DeleteContext(ContextGL* context)
{
  ContextGLX* ctx = (ContextGLX*)context;
  glXDestroyContext(mWinHandle.dpy,ctx->glXContext );
  delete ctx;
  
  mCurrentContext = NULL;
}


void WindowGL::SetContext(ContextGL* context)
{
  ContextGLX* ctx = (ContextGLX*)context;

  if ( ctx )
    glXMakeCurrent(mWinHandle.dpy, mWinHandle.win,ctx->glXContext);
  
  mCurrentContext = context;

  cout << "GL_VENDOR: " << glGetString(GL_VENDOR) << endl << endl;
  cout << "GL_RENDERER: " << glGetString(GL_RENDERER) << endl << endl;
  cout << "GL_VERSION: " << glGetString(GL_VERSION) << endl << endl;
  cout << "GL_EXTENSIONS: " << glGetString(GL_EXTENSIONS) << endl << endl;

}

void WindowGL::PageFlip(bool retrace)
{
  ContextGLX* ctx = (ContextGLX*)mCurrentContext;
  if ( ctx )
    {
      glXSwapBuffers(mWinHandle.dpy, mWinHandle.win); 
    }  
}
