/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source:
		memorystream implementation

	revision history:
		Mar/19/2001 - Jukka Liimatta - initial revision
*/
#include <prcore/filestream.hpp>
#include <prcore/memorystream.hpp>

using namespace prcore;


MemoryStream::MemoryStream()
: mOffset(0), mSize(0), mMaxSize(0), mData(NULL)
{
}


MemoryStream::MemoryStream(int size)
: mOffset(0), mSize(size), mMaxSize(size), mData(NULL)
{
	mData = new uint8[ size ];
}


MemoryStream::MemoryStream(uint8* data, int size)
: mOffset(0), mSize(size), mMaxSize(size), mData(data)
{
}


MemoryStream::MemoryStream(const char* filename)
: mOffset(0), mSize(0), mMaxSize(0), mData(NULL)
{
	FileStream file(filename);

	mSize = file.GetSize();
	mMaxSize = mSize;
	mData = new uint8[ mSize ];
	file.Read(mData,mSize);
}


MemoryStream::~MemoryStream()
{
	if ( mData )
		delete[] mData;
}


int MemoryStream::Seek(int delta, SeekMode mode)
{
	switch( mode )
	{
		case START:		mOffset = delta; break;
		case CURRENT:	mOffset += delta; break;
		case END:		mOffset = mSize - delta; break;
	}
	return mOffset;
}


int MemoryStream::Read(void* target, int bytes)
{
	int overflow = mOffset + bytes - mSize;//todo
	if ( overflow > 0 )
	{
		bytes -= overflow;
		if ( bytes < 1 ) 
			return 0;
	}

	memcpy(target,mData+mOffset,bytes);
	mOffset += bytes;
	return bytes;
}


int MemoryStream::Write(const void* source, int bytes)
{
	if ( bytes < 1 )
		return 0;

	// overflow?
	if ( (mOffset+bytes) > mMaxSize )
	{
		int size = (int)(mMaxSize * 1.25 + bytes);
		uint8* data = new uint8[ size ];

		if ( mData )
		{
			memcpy(data,mData,mSize);
			delete[] mData;
		}

		mMaxSize = size;
		mData = data;
	}

	// append source data
	memcpy(mData+mOffset,source,bytes);
	mOffset += bytes;

	// adjust size
	if ( mOffset > mSize )
		mSize = mOffset;

	return bytes;
}


int MemoryStream::GetOffset() const
{
	return mOffset;
}


int MemoryStream::GetSize() const
{
	return mSize;
}


bool MemoryStream::IsOpen() const
{
	return true;
}


bool MemoryStream::IsEOS() const
{
	return mOffset >= mSize;
}


uint8* MemoryStream::GetData() const
{
	return mData;
}

