/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source:
		string

	revision history:
		Jun/10/2000 - Thomas Mannfred Carlsson - initial revision
		Jan/24/2001 - Jukka Liimatta - renaissance build
		Nov/16/2001 - Mats Byggmastar - reported a memory leak in "+ operator"(s)
		Dec/01/2001 - Jukka Liimatta - rewrote the string implementation
		Feb/06/2002 - Francesco Montecuccoli - added SubString()
		
	note:
	- optimized for size, not speed (efficiency of +=, << operators in general is not optimal)
	- ie. construction is slow, storage is optimal
*/
#include <stdarg.h>
#include <prcore/string.hpp>

using namespace prcore;


// ===========================================
// prostring()
// ===========================================

static inline int prolen(const char* text)
{
	if ( !text )
		return 0;

	const char* s;

	for ( s = text; *s; s++ );
	return (int)(s - text);
}

static inline void procpy(char* d, const char* s, int count)
{
	while ( count-- )
		*d++ = *s++;
}

static inline void procpy(char* d, const char* s1, int count1, const char* s2, int count2)
{
	if ( s1 )
	{
		while ( count1-- )
			*d++ = *s1++;
	}

	if ( s2 )
	{
		while ( count2-- )
			*d++ = *s2++;
	}

	// terminate string
	*d = 0;
}


// ===========================================
// String
// ===========================================

String::String()
: mData(NULL),mLength(0)
{
}


String::String(const String& s)
: mData(NULL),mLength(0)
{
	*this = s;
}


String::String(const char* text)
: mData(NULL),mLength(0)
{
	*this = text;
}


String::String(char* text, int length)
: mData(text),mLength(length)
{
}


String::~String()
{
	delete[] mData;
}


String& String::operator << (const short s)
{
	char text[8];
	sprintf(text,"%d",(int)s);
	return operator += (text);
}


String& String::operator << (const unsigned short us)
{
	char text[8];
	sprintf(text,"%u",(unsigned int)us);
	return operator += (text);
}


String& String::operator << (const int i)
{
	char text[16];
	sprintf(text,"%d",(int)i);
	return operator += (text);
}


String& String::operator << (const unsigned int ui)
{
	char text[16];
	sprintf(text,"%u",(unsigned int)ui);
	return operator += (text);
}


String& String::operator << (const long l)
{
	char text[16];
	sprintf(text,"%d",(int)l);
	return operator += (text);
}


String& String::operator << (const unsigned long ul)
{
	char text[16];
	sprintf(text,"%u",(unsigned int)ul);
	return operator += (text);
}


String& String::operator << (const float f)
{
	PrintText("%.3f",f);
	return *this;
}


String& String::operator = (const String& s)
{
	delete[] mData;

	if ( s.mData )
	{
		int size = s.mLength;
		mData = new char[size+1];
		mLength = size;

		procpy(mData,s.mData,size+1);
	}
	else
	{
		mData = NULL;
		mLength = 0;
	}

	return *this;
}


String& String::operator = (const char* text)
{
	delete[] mData;

	if ( text )
	{
		int size = prolen(text);
		mData = new char[size+1];
		mLength = size;

		procpy(mData,text,size+1);
	}
	else
	{
		mData = NULL;
		mLength = 0;
	}

	return *this;
}


String& String::operator += (const String& s)
{
	int size = mLength + s.mLength;
	char* data = NULL;

	if ( size )
	{
		data = new char[size+1];
		procpy(data,mData,mLength,s.mData,s.mLength);
	}

	delete[] mData;
	mData = data;
	mLength = size;

	return *this;
}


String& String::operator += (const char* text)
{
	int textsize = prolen(text);
	int size = mLength + textsize;
	char* data = NULL;

	if ( size )
	{
		data = new char[size+1];
		procpy(data,mData,mLength,text,textsize);
	}

	delete[] mData;
	mData = data;
	mLength = size;

	return *this;
}


String String::operator + (const String& s) const
{
	int size = mLength + s.mLength;
	char* data = NULL;

	if ( size )
	{
		data = new char[size+1];
		procpy(data,mData,mLength,s.mData,s.mLength);
	}

	return String(data,size);
}


String String::operator + (const char* text) const
{
	int textsize = prolen(text);
	int size = mLength + textsize;
	char* data = NULL;

	if ( size )
	{
		data = new char[size+1];
		procpy(data,mData,mLength,text,textsize);
	}

	return String(data,size);
}


bool String::operator == (const String& s) const
{
	if ( (mLength != s.mLength) || !mData || !s.mData )
		return false;

	const char* s1 = mData;
	const char* s2 = s.mData;
	int count = s.mLength;

	while ( count-- )
	{
		if ( *s1++ != *s2++ )
			return false;
	}
	return true;
}


bool String::operator == (const char* text) const
{
	if ( !mData || !text )
		return false;

	const char* s1 = mData;

	while ( *s1 == *text++ )
	{
		if ( *s1++ == 0 )
			return true;
	}
	return false;
}


bool String::operator != (const String& s) const
{
	if ( (mLength != s.mLength) || !mData || !s.mData )
		return true;

	const char* s1 = mData;
	const char* s2 = s.mData;
	int count = s.mLength;

	while ( count-- )
	{
		if ( *s1++ == *s2++ )
			return false;
	}
	return true;
}


bool String::operator != (const char* text) const
{
	if ( !mData || !text )
		return true;

	const char* s1 = mData;
	const char* s2 = text;

	while ( *s1 == *s2++ )
	{
		if ( *s1++ == 0 )
			return false;
	}
	return true;
}


void String::PrintText(const char* text, ...)
{
	assert( text );

    // WARNING: this buffer is overflowable - do not use PrintText() as-is
	// in a security-critical environment for blind processing of user input.
	char buffer[256];
    va_list args;

	va_start(args,text);
    vsprintf(buffer,text,args);
    va_end(args);

	operator += (buffer);
}

String String::SubString(int offset, int length) const
{
	if ( offset < 0 || offset >= mLength || length <= 0 )
		return "";

	int size = (offset + length > mLength ? mLength - offset : length);

	char* data = new char[size+1];
	procpy(data,mData+offset,size);
	data[size] = 0;

	return String(data,size);
}
