/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source:
		win32 mouse

	revision history:
		Jan/07/2000 - Thomas Mannfred Carlsson - initial revision
		Jan/24/2001 - Jukka Liimatta - renaissance build

	todo:
	- Mouse velocity/axis implemenation
	- Change behaviour to work "correctly" in windowed mode
*/
#include <prcore/configure.hpp>
#include <prcore/mouse.hpp>

using namespace prcore;


// =================================================
// statics
// =================================================

static LRESULT CALLBACK mouseStaticHook(int ncode, WPARAM wparam, LPARAM lparam);

static int mouseFlags = 0;
static int mouseX = 0;
static int mouseY = 0;
static HHOOK mouseStaticHookId = 0;
static long mouseStaticInitCount = 0;


// =================================================
// Mouse
// =================================================

Mouse::Mouse()
{
	mouseStaticInitCount++;

	if ( ( !mouseStaticHookId ) && ( mouseStaticInitCount <= 1 ) )
	{
		mouseStaticHookId = SetWindowsHookEx( WH_MOUSE,
			(HOOKPROC) mouseStaticHook, 0, GetCurrentThreadId() );

		if ( !mouseStaticHookId ) 
			return;
	}
}


Mouse::~Mouse()
{
	if ( mouseStaticInitCount > 0 ) mouseStaticInitCount--;

	if ( ( mouseStaticHookId ) && ( mouseStaticInitCount == 0 ) )
	{
		UnhookWindowsHookEx( mouseStaticHookId );
		mouseStaticHookId = 0;

		ClipCursor( NULL );	// restore when done
	}
}


void Mouse::SetArea(const Rect& rect)
{
	RECT clip = 
	{ 
		rect.x,
		rect.y,
		rect.x+rect.width-1,
		rect.y+rect.height-1
	};
	ClipCursor( &clip );
}


void Mouse::SetXY(int x, int y)
{
	SetCursorPos( x, y );
}


int Mouse::GetX()
{
	return mouseX;
}


int Mouse::GetY()
{
	return mouseY;
}


int Mouse::GetAxisCount() const
{
	return 0;
}


int Mouse::GetButtonCount() const
{
	return 3;
}


float Mouse::GetAxis(int index)
{
	return 0;
}


bool Mouse::GetButton(int index)
{
	switch ( index )
	{
		case MOUSE_LEFT:	return (mouseFlags & MOUSE_LEFT) != 0;
		case MOUSE_RIGHT:	return (mouseFlags & MOUSE_RIGHT) != 0;
		case MOUSE_MIDDLE:	return (mouseFlags & MOUSE_MIDDLE) != 0;
		default:			return false;
	}
}


static LRESULT CALLBACK mouseStaticHook( int ncode, WPARAM wparam, LPARAM lparam )
{
	if ( ncode < 0 )
	{
		return (int) CallNextHookEx( mouseStaticHookId, ncode, wparam, lparam );
	}

	mouseX = ( (MOUSEHOOKSTRUCT *)lparam )->pt.x;
	mouseY = ( (MOUSEHOOKSTRUCT *)lparam )->pt.y;

	switch ( wparam )
	{
		case WM_LBUTTONDOWN:
		{
			mouseFlags |= MOUSE_LEFT;
			break;
		}
		case WM_LBUTTONUP:
		{
			mouseFlags &= ( 0x0ffffff - MOUSE_LEFT );
			break;
		}
		case WM_MBUTTONDOWN:
		{
			mouseFlags |= MOUSE_MIDDLE;
			break;
		}
		case WM_MBUTTONUP:
		{
			mouseFlags &= ( 0x0ffffff - MOUSE_MIDDLE );
			break;
		}
		case WM_RBUTTONDOWN:
		{
			mouseFlags |= MOUSE_RIGHT;
			break;
		}
		case WM_RBUTTONUP:
		{
			mouseFlags &= ( 0x0ffffff - MOUSE_RIGHT );
			break;
		}
	}

	return 1;
}

