/*
	Twilight Prophecy SDK
	A multi-platform development system for virtual reality and multimedia.

	Copyright (C) 1997-2002 Twilight 3D Finland Oy Ltd.


    This software is provided 'as-is', without any express or implied
    warranty.  In no event will the authors be held liable for any damages
    arising from the use of this software.

    Permission is granted to anyone to use this software for any purpose,
    including commercial applications, and to alter it and redistribute it
    freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not
       claim that you wrote the original software. If you use this software
       in a product, an acknowledgment in the product documentation would be
       appreciated but is not required.
    2. Altered source versions must be plainly marked as such, and must not be
       misrepresented as being the original software.
    3. This notice may not be removed or altered from any source distribution.


	source: 
		intersection implementation

	revision history:
		Jan/31/2001 - Jukka Liimatta - initial revision
		Nov/30/2001 - Jukka Liimatta - improved IntersectRayBox(), minor fixes
*/
#define PRMATH_TYPENAME
#include <prmath/prmath.hpp>
#include <float.h>

using namespace prmath;


bool prmath::IntersectConeSphere(const Cone& cone, const Sphere& sphere)
{
	// test if cone vertex is in sphere
	Vector3 diff = sphere.center - cone.origin;
	
	float r2 = sphere.radius * sphere.radius;
	float len2 = DotProduct(diff,diff);
	float len2mr2 = len2 - r2;
	if ( len2mr2 <= 0 ) 
		return true;

	// test if sphere center is in cone
	float dot1 = DotProduct(cone.target,diff);
	float dot2 = dot1 * dot1;
	float cs = (float)cos(cone.angle);
	float cs2 = cs * cs;
	if ( dot2 >= len2*cs2 ) 
		return true;
	
	float ulen = (float)sqrt(fabs(len2-dot2));
	float sn = (float)sin(cone.angle);
	float test = cs*dot1 + sn*ulen;
	float discr = test*test - len2mr2;
	
	return discr >= 0 && test >= 0;
}


bool prmath::IntersectBoxSphere(const Box& box, const Sphere& sphere)
{
	float fmin = 0;
	for ( int i=0; i<3; i++ )
	{
		const float& center = sphere.center[i];
		const float& min = box.vmin[i];
		const float& max = box.vmax[i];

		if (center < min)
		{
			float delta = center - min;
			fmin += delta*delta;
		}
		else if (center > max)
		{
			float delta = center - max;
			fmin += delta*delta;
		}
	}
	float r2 = sphere.radius * sphere.radius;
	return (fmin <= r2) ? true : false;
}


bool prmath::IntersectRaySphere(const Line& ray, const Sphere& sphere, float& enter)
{
	Vector3 dep = sphere.center - ray.origin;
	float dot3 = DotProduct(dep, ray.direction);
	float det = dot3*dot3 - LengthSquared(dep) + sphere.radius*sphere.radius;

	// no intersection
	if ( det < 0 )
		return false;

	// positive solution, or tangent (det==0)
	det = (float)sqrt(det);
	float t = dot3 + det;

	// intersection behind the origin for positive solution?
	if ( t < 0 )
	{
		// try negative solution, since the positive was behind the origin
		t = dot3 - det;
		if ( t < 0 ) 
			return false; // also behind origin
	}
	else
	{
		// negative solution
		if ( det > 0 )
		{
			float to = dot3 - det;
			// use negative solution only if closer to origin, and not behind it
			if ( (to < t) && (to >= 0) ) 
				t = to;
		}
	}

	// intersection result
	enter = t;

	return true;
}


bool prmath::IntersectRaySphere(const Line& ray, const Sphere& sphere, float& enter, float& leave)
{
	Vector3 direction = Normalize(ray.direction);
	float b = 2 * (DotProduct(ray.origin,direction) - DotProduct(sphere.center,direction));
	float c = DotProduct(ray.origin,ray.origin) +	DotProduct(sphere.center,sphere.center) - 2 * DotProduct(ray.origin,sphere.center) - sphere.radius*sphere.radius;
		
	float det = b * b - 4 * c;
	if ( det >= 0 )
	{
		float sd = (float)sqrt(det);
		float t0 = (-b + sd) * 0.5f;
		float t1 = (-b - sd) * 0.5f;
		if ( t0 > t1 )
		{
			enter = t1;
			leave = t0;
		}
		else
		{
			enter = t0;
			leave = t1;
		}
		return true;
	}
	return false;
}


bool prmath::IntersectRayPlane(const Line& ray, const Plane& plane, float& enter)
{
	float ndot = (DotProduct(ray.origin,plane.normal) + plane.dist);
	if ( ndot > 0 )
		return false;
		
	// intersection result
	enter = ndot / DotProduct(ray.direction,plane.normal);

	return true;
}


bool prmath::IntersectRayTriangle(const Line& ray, const Vector3& v0, const Vector3& v1, const Vector3& v2, float& enter)
{
	// intersect the ray with plane
	plane3f plane(v0,v1,v2);
	if ( !IntersectRayPlane(ray,plane,enter) )
		return false;

	// intersection point
	point3f point = ray.origin + ray.direction * enter;

	// test if intersection point is inside the triangle
	const vec3f& origin = ray.origin;
	vec3f normal;
	
	// edge 1
	normal = CrossProduct(origin-v1,origin-v0);
	if ( DotProduct(point,normal) < DotProduct(origin,normal) )
		return false;
		
	// edge 2
	normal = CrossProduct(origin-v2,origin-v1);
	if ( DotProduct(point,normal) < DotProduct(origin,normal) )
		return false;
		
	// edge 3
	normal = CrossProduct(origin-v0,origin-v2);
	if ( DotProduct(point,normal) < DotProduct(origin,normal) )
		return false;
		
	return true;
}


bool prmath::IntersectRayBox(const Line& ray, const Box& box, float& enter, float& leave)
{
	// =====================================
	// Kay and Kayjia "slab" method
	// =====================================
	
	float tnear = -FLT_MAX;
	float tfar = FLT_MAX;
	
	for ( int i=0; i<3; i++ )
	{
		float origin = ray.origin[i];
		float direction = ray.direction[i];
		float min = box.vmin[i];
		float max = box.vmax[i];
		
		if ( direction )
		{
			// calculate intersections to the slab
			direction = 1.0f / direction;
			float t1 = (min - origin) * direction;
			float t2 = (max - origin) * direction;
			
			if ( t1 > t2 )
			{
				if ( t2 > tnear ) tnear = t2;
				if ( t1 < tfar ) tfar = t1;
			}
			else
			{
				if ( t1 > tnear ) tnear = t1;
				if ( t2 < tfar ) tfar = t2;
			}
		
			// failed intersecting this axis or intersecting behind the origin
			if ( tnear > tfar || tfar < 0 )
				return false;
		}
		else
		{
			// ray parallel to axis/slab being tested against
			if ( origin < min || origin > max )
				return false;
		}
	}

	// intersection result
	enter = tnear;
	leave = tfar;

	// ray hits box
	return true;
}	


bool prmath::PlaneIntersect(const Plane& p0, const Plane& p1, Vector3& point, Vector3& direction)
{
	direction = CrossProduct(p0.normal,p1.normal);
	float det = LengthSquared(direction);
	
	// parallel?
	if ( det < (float)0.000001 ) 
		return false;

	// intersection
	point = (p0.dist * CrossProduct(p1.normal,direction) + p1.dist * CrossProduct(direction,p0.normal)) / det;
	direction = Normalize(direction);

	return true;
}


bool prmath::PlaneIntersect(const Plane& p0, const Plane& p1, const Plane& p2, Vector3& point)
{
	// determinant
	Vector3 cp01 = CrossProduct(p0.normal,p1.normal);
	float det = DotProduct(cp01,p2.normal);
	
	// parallel?
	if ( det < (float)0.001 ) 
		return false;

	// intersection
	Vector3 cp12 = CrossProduct(p1.normal,p2.normal) * p0.dist;
	Vector3 cp20 = CrossProduct(p2.normal,p0.normal) * p1.dist;

	point = (cp12 + cp20 + p2.dist * cp01) / det;
	return true;
}


bool prmath::PointInsideTriangle(const Vector3& v0, const Vector3& v1, const Vector3& v2, const Vector3& point)
{
	Vector3 v10 = v1 - v0;
	Vector3 v20 = v2 - v0;
	Vector3 vp0 = point - v0;
	float b = DotProduct(v10,v20);
	float c = DotProduct(v20,v20);
	float d = DotProduct(vp0,v10);
	float e = DotProduct(vp0,v20);
	
	float x = d*c - e*b;
	if ( x < 0 )
		return false;

	float a = DotProduct(v10,v10);
	float y = e*a - d*b;
	if ( y < 0 )
		return false;

	return (x+y) <= (a*c - b*b);
}


float prmath::GetTriangleArea(float edge0, float edge1, float edge2)
{
	// Heron's formula
	float s = (edge0 + edge1 + edge2) * 0.5f;
	return (float)sqrt( s * (s-edge0) * (s-edge1) * (s-edge2) );
}

